///File : Image.cpp
//------------------------------------------------------------------------------
/*
    This file is part of DevisMaker.

    DevisMaker is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    DevisMaker is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with DevisMaker.  If not, see <http://www.gnu.org/licenses/>.
*/
//------------------------------------------------------------------------------
#include "Sizer.hpp"
//------------------------------------------------------------------------------

namespace mdo
{

//------------------------------------------------------------------------------
//---------------------------------- Sizer -------------------------------------
//------------------------------------------------------------------------------
bool Sizer::CalculSizeReal(wxDC *dc, const int &width, const int &askedHeight)
{
    std::vector<wxSize> mdoSizeTbl;
    mdoSizeTbl.resize(m_data.mdoList.size());


    // On calcule les tailles des mdos...
    m_isReadyToDraw = true;
    if(m_data.isVertical) // vertical
    {
        // --------------- WIDTH ---------------
        for(int i = 0; i < m_data.mdoList.size(); i++)
        {
            int mdoWidth;
            if(m_data.mdoList[i]->GetWidthProportion() == -1)
            {
                mdoWidth = width; // Il veut tout ==> tout est à lui
            }
            else
            {
                mdoWidth = width * m_data.mdoList[i]->GetWidthProportion();
            }

            mdoSizeTbl[i].SetWidth(mdoWidth);
        }

        // --------------- HEIGHT ---------------
        // On commence par ceux qui ont une height definie...
        int leftMdoHeight = askedHeight; // Espace utilisé par les mdo à taille definie
        int numberOfExpandMdo = 0; // nombre de mdo ayant une height de -1 (le max possible) pour partage
        for(int i = 0; i < m_data.mdoList.size(); i++)
        {
            if(m_data.mdoList[i]->GetHeightProportion() != -1)
            {
                int currentMdoHeight = askedHeight * m_data.mdoList[i]->GetHeightProportion();
                mdoSizeTbl[i].SetHeight(currentMdoHeight);
                leftMdoHeight -= currentMdoHeight;
            }
            else
            {
                numberOfExpandMdo++;
            }
        }

        // Pour finir : ceux qui veulent le max ("expand")
        if(numberOfExpandMdo > 0)
        {
            int expandMdoHeight;
            if(leftMdoHeight < 0) // on met à 1 si c'est negatif
            {
                expandMdoHeight = 1;
            }
            else
            {
                expandMdoHeight = leftMdoHeight / numberOfExpandMdo;
            }

            for(int i = 0; i < m_data.mdoList.size(); i++)
            {
                if(m_data.mdoList[i]->GetHeightProportion() == -1)
                {
                    mdoSizeTbl[i].SetHeight(expandMdoHeight);
                }
            }
        }
    }
    else // horizontal : partage de la width
    {
        // --------------- WIDTH ---------------
        // On commence par ceux qui ont une width definie...
        int leftMdoWidth = width; // Espace utilisé par les mdo à taille definie
        int numberOfExpandMdo = 0; // nombre de mdo ayant une width de -1 (le max possible) pour partage
        for(int i = 0; i < m_data.mdoList.size(); i++)
        {
            if(m_data.mdoList[i]->GetWidthProportion() != -1)
            {
                int currentMdoWidth = width * m_data.mdoList[i]->GetWidthProportion();
                mdoSizeTbl[i].SetWidth(currentMdoWidth);
                leftMdoWidth -= currentMdoWidth;
            }
            else
            {
                numberOfExpandMdo++;
            }
        }

        // Pour finir : ceux qui veulent le max ("expand")
        if(numberOfExpandMdo > 0)
        {
            int expandMdoWidth;
            if(leftMdoWidth < 0) // on met à 1 si c'est negatif
            {
                expandMdoWidth = 1;
            }
            else
            {
                expandMdoWidth = leftMdoWidth / numberOfExpandMdo;
            }

            for(int i = 0; i < m_data.mdoList.size(); i++)
            {
                if(m_data.mdoList[i]->GetWidthProportion() == -1)
                {
                    mdoSizeTbl[i].SetWidth(expandMdoWidth);
                }
            }
        }

        // --------------- HEIGHT ---------------
        for(int i = 0; i < m_data.mdoList.size(); i++)
        {
            int mdoHeight;
            if(m_data.mdoList[i]->GetHeightProportion() == -1.f)
            {
                mdoHeight = askedHeight; // Il veut tout ==> tout est à lui
            }
            else
            {
                mdoHeight = askedHeight * m_data.mdoList[i]->GetHeightProportion();
            }

            mdoSizeTbl[i].SetHeight(mdoHeight);
        }
    }

    // On applique aux mdos réélement...
    for(int i = 0; i < m_data.mdoList.size(); i++)
    {
        m_data.mdoList[i]->CalculSizeReal(dc, mdoSizeTbl[i].GetWidth(), mdoSizeTbl[i].GetHeight());
    }

    // Calcul de la taille du sizer
    int sizerWidth = 0, sizerHeight = 0;
    if( m_data.isVertical ) // VERTICAL
    {
        for(int i = 0; i < m_data.mdoList.size(); i++)
        {
            wxSize currentSize = m_data.mdoList[i]->GetSizeReal();

            // width : max
            sizerWidth = std::max(sizerWidth, currentSize.GetWidth());

            // height : addition
            sizerHeight += currentSize.GetHeight();
        }
    }
    else // HORIZONTAL
    {
        for(int i = 0; i < m_data.mdoList.size(); i++)
        {
            wxSize currentSize = m_data.mdoList[i]->GetSizeReal();

            // height : max
            sizerHeight = std::max(sizerHeight, currentSize.GetHeight());

            // width : addition
            sizerWidth += currentSize.GetWidth();
        }

        // CAS SPECIAL : SI LE SIZER EST AGRANDI, IL FAUT RECALCULER LES HAUTEURS...
        if(sizerHeight > askedHeight)
        {
            // --------------- HEIGHT ---------------
            for(int i = 0; i < m_data.mdoList.size(); i++)
            {
                int mdoHeight;
                if(m_data.mdoList[i]->GetHeightProportion() == -1.f)
                {
                    mdoHeight = sizerHeight; // Il veut tout ==> tout est à lui
                }
                else
                {
                    mdoHeight = sizerHeight * m_data.mdoList[i]->GetHeightProportion();
                }

                mdoSizeTbl[i].SetHeight(mdoHeight);
            }

            // On applique aux mdos réélement...
            for(int i = 0; i < m_data.mdoList.size(); i++)
            {
                m_data.mdoList[i]->CalculSizeReal(dc, mdoSizeTbl[i].GetWidth(), mdoSizeTbl[i].GetHeight());
            }
        }
    }

    // On met au max (si les enfants font moins de taille que prévu
    sizerWidth = std::max(sizerWidth, width);
    sizerHeight = std::max(sizerHeight, askedHeight);

    m_sizeReal.SetWidth(sizerWidth);
    m_sizeReal.SetHeight(sizerHeight);
}
//------------------------------------------------------------------------------
Sizer::Sizer(ConfigManager *cfg, ConfigManager *devisCfg, DrawableObject *parrentSizer, const std::string &name) : DrawableObject(cfg, devisCfg, parrentSizer, name)
{
    m_typeName = MDO_SIZER_TYPE_NAME;
    SetConfigName(""); // Rechargement du prefix
}
//------------------------------------------------------------------------------
void Sizer::LoadFromCfg(const std::string &name)
{
    ClearMdos();

    DrawableObject::LoadFromCfg(name);
    m_data.isBorder = m_cfg->GetBool(m_cfgPrefix + "isBorder");
    m_data.mdoStrList = m_cfg->GetString(m_cfgPrefix + "mdoStrList");
    m_data.isVertical = m_cfg->GetBool(m_cfgPrefix + "isVertical");

    size_t lastPos = 0;
    size_t pos = m_data.mdoStrList.find(';', lastPos);
    while(pos != std::string::npos)
    {
        std::string currentMdo = m_data.mdoStrList.substr(lastPos, pos - lastPos);

        size_t separatorPos = currentMdo.find(':');
        if(separatorPos != std::string::npos)
        {
            std::string type = currentMdo.substr(0, separatorPos);
            std::string name = currentMdo.substr(separatorPos + 1);

            LoadMdo(type, name);
        }

        lastPos = pos + 1;
        pos = m_data.mdoStrList.find(';', lastPos);
    }
}
//------------------------------------------------------------------------------
void Sizer::SaveToCfg(const std::string &name)
{
    DrawableObject::SaveToCfg(name);
    m_cfg->SetBool(m_cfgPrefix + "isBorder", m_data.isBorder);
    m_cfg->SetString(m_cfgPrefix + "mdoStrList", m_data.mdoStrList);
    m_cfg->SetBool(m_cfgPrefix + "isVertical", m_data.isVertical);
}
//------------------------------------------------------------------------------
void Sizer::CreateToCfg(const std::string &name)
{
    DrawableObject::CreateToCfg(name);
    m_cfg->SetBool(m_cfgPrefix + "isBorder", false);
    m_cfg->SetString(m_cfgPrefix + "mdoStrList", "");
    m_cfg->SetBool(m_cfgPrefix + "isVertical", false);

    LoadFromCfg();
}
//------------------------------------------------------------------------------
void Sizer::RemoveFromCfg(const std::string &name, const bool &withChilds)
{
    if(withChilds)
    {
        while( m_data.mdoList.size() != 0)
        {
            EraseMdo(0);
        }
    }

    DrawableObject::RemoveFromCfg(name);
    m_cfg->RemoveBool(m_cfgPrefix + "isBorder");
    m_cfg->RemoveString(m_cfgPrefix + "mdoStrList");
    m_cfg->RemoveBool(m_cfgPrefix + "isVertical");
}
//------------------------------------------------------------------------------
void Sizer::EditPropertyDialog(DialogProperty *window)
{
    DrawableObject::EditPropertyDialog(window);

    wxPanel *PagePanel;
    wxBoxSizer *PageSizer;

    // Special Mdo
    PagePanel = new wxPanel(window->GetNotebook(), wxID_ANY);
    PageSizer = new wxBoxSizer(wxVERTICAL);

    window->AddParamCtrl(type_check, PagePanel, PageSizer, m_cfgPrefix + "isBorder", _("Add border"));
    window->AddParamCtrl(type_check, PagePanel, PageSizer, m_cfgPrefix + "isVertical", _("This sizer is vertical"));

    PagePanel->SetSizer(PageSizer);
    window->GetNotebook()->AddPage(PagePanel, _("Object"));
}
//------------------------------------------------------------------------------
void Sizer::Draw(wxDC *dc, const wxPoint &pos)
{
    ThrowIfNotReadyToDraw();

    wxPoint nextPos = pos;

    if( m_data.isVertical ) // nextPos.y augmente.
    {
        for(int i = 0; i < m_data.mdoList.size(); i++)
        {
            m_data.mdoList[i]->Draw(dc, nextPos);
            nextPos.y += m_data.mdoList[i]->GetSizeReal().GetY();
        }
    }
    else // nextPos.x augmente.
    {
        for(int i = 0; i < m_data.mdoList.size(); i++)
        {
            m_data.mdoList[i]->Draw(dc, nextPos);
            nextPos.x += m_data.mdoList[i]->GetSizeReal().GetX();
        }
    }

    if(m_data.isBorder)
    {
        // Bordures
        dc->DrawRoundedRectangle(pos.x, pos.y, GetSizeReal().GetWidth(), GetSizeReal().GetHeight(), m_cfg->GetInt("arcSize"));
    }
}
//------------------------------------------------------------------------------
Sizer::~Sizer()
{
    ClearMdos();
}
//------------------------------------------------------------------------------

// Special Sizer :
void Sizer::ClearMdos()
{
    while( ! m_data.mdoList.empty())
    {
        delete m_data.mdoList[m_data.mdoList.size() - 1];
        m_data.mdoList.pop_back();
    }

    CreateMdoListStr();
}
//------------------------------------------------------------------------------
void Sizer::EraseMdo(const int &pos)
{
    // Data
    if( DrawableObjectExistsInCfg(m_cfg, m_data.mdoList[pos]->GetType(), m_data.mdoList[pos]->GetConfigName() )) // On supprime de la config que si c'est nécessaire
    {
        m_data.mdoList[pos]->RemoveFromCfg();
    }

    // Suppression de l'objet
    delete m_data.mdoList[pos];
    m_data.mdoList.erase(m_data.mdoList.begin() + pos);

    CreateMdoListStr();
    SaveToCfg();
}
//------------------------------------------------------------------------------
void Sizer::EraseMdo(DrawableObject *mdo)
{
    for(int i = 0; i < m_data.mdoList.size(); i++) // Pas de "break;" par plusieurs pointeur possible.
    {
        if(mdo == m_data.mdoList[i])
        {
            EraseMdo(i);
        }
    }

    CreateMdoListStr();
    SaveToCfg();
}
//------------------------------------------------------------------------------
/*void Sizer::RenameMdo(DrawableObject *mdo, const std::string &newName)
{
    mdo->RemoveFromCfg("", false);
    mdo->SetConfigName(newName);
    mdo->SaveToCfg();
    CreateMdoListStr();
    SaveToCfg();
}*/
//------------------------------------------------------------------------------
void Sizer::ReceiveMdo(DrawableObject *mdo, const DrawableObject *after, const bool &addAfterDest)
{
    if(after == NULL)
    {
        m_data.mdoList.push_back(mdo);
    }
    else
    {
        int pos = GetMdoPos(after);
        if(addAfterDest) // On se met après et pas à cette place...
        {
            pos ++;
        }
        m_data.mdoList.insert(m_data.mdoList.begin() + pos, mdo);
    }

    mdo->SetParent(this); // Adoptation
    CreateMdoListStr();
    SaveToCfg();
}
//------------------------------------------------------------------------------
void Sizer::RefreshParentLink()
{
    for(int i = 0; i < m_data.mdoList.size(); i++) // Pas de "break;" par plusieurs pointeur possible.
    {
        m_data.mdoList[i]->SetParent(this);

    }
}
//------------------------------------------------------------------------------
void Sizer::CreateOwnCopy()
{
    for(int pos = 0; pos < m_data.mdoList.size(); pos++)
    {
        if(m_data.mdoList[pos]->GetType() == MDO_STATIC_TEXT_TYPE_NAME)
        {
            m_data.mdoList[pos] = new StaticText(*static_cast<StaticText*>(m_data.mdoList[pos]));
        }
        else if(m_data.mdoList[pos]->GetType() == MDO_DEVIS_TEXT_TYPE_NAME)
        {
            m_data.mdoList[pos] = new DevisText(*static_cast<DevisText*>(m_data.mdoList[pos]));
        }
        else if(m_data.mdoList[pos]->GetType() == MDO_IMAGE_TYPE_NAME)
        {
            m_data.mdoList[pos] = new Image(*static_cast<Image*>(m_data.mdoList[pos]));
        }
        else if(m_data.mdoList[pos]->GetType() == MDO_SIZER_TYPE_NAME)
        {
            Sizer *newSizer = new Sizer(*static_cast<Sizer*>(m_data.mdoList[pos]));
            m_data.mdoList[pos] = newSizer;

            newSizer->CreateOwnCopy();
            newSizer->RefreshParentLink();
        }
        else if(m_data.mdoList[pos]->GetType() == MDO_SPACER_TYPE_NAME)
        {
            m_data.mdoList[pos] = new Spacer(*static_cast<Spacer*>(m_data.mdoList[pos]));
        }
        else
        {
            std::string msg = "mdo::Sizer CreateOwnCopy ==> Unknown mdo mdo->GetType() \"" + m_data.mdoList[pos]->GetType() + "\".";
            throw ticpp::Exception(msg.c_str());
        }

        m_data.mdoList[pos]->AutoChangeConfigName();
        m_data.mdoList[pos]->SaveToCfg();
    }
}
//------------------------------------------------------------------------------

void Sizer::CopyMdo(DrawableObject *mdo, const DrawableObject *after, const bool &addAfterDest)
{
    DrawableObject *cpMdo;
    if(mdo->GetType() == MDO_STATIC_TEXT_TYPE_NAME)
    {
        cpMdo = new StaticText(*static_cast<StaticText*>(mdo));
    }
    else if(mdo->GetType() == MDO_DEVIS_TEXT_TYPE_NAME)
    {
        cpMdo = new DevisText(*static_cast<DevisText*>(mdo));
    }
    else if(mdo->GetType() == MDO_IMAGE_TYPE_NAME)
    {
        cpMdo = new Image(*static_cast<Image*>(mdo));
    }
    else if(mdo->GetType() == MDO_SIZER_TYPE_NAME)
    {
        Sizer *newSizer = new Sizer(*static_cast<Sizer*>(mdo));
        cpMdo = newSizer;

        newSizer->CreateOwnCopy();
        newSizer->RefreshParentLink();
    }
    else if(mdo->GetType() == MDO_SPACER_TYPE_NAME)
    {
        cpMdo = new Spacer(*static_cast<Spacer*>(mdo));
    }
    else
    {
        std::string msg = "mdo::Sizer CopyMdo ==> Unknown mdo mdo->GetType() \"" + mdo->GetType() + "\".";
        throw ticpp::Exception(msg.c_str());
    }

    // Chagement de nom
    cpMdo->AutoChangeConfigName();

    if(after == NULL)
    {
        m_data.mdoList.push_back(cpMdo);
    }
    else
    {
        int pos = GetMdoPos(after);
        if(addAfterDest) // On se met après et pas à cette place...
        {
            pos ++;
        }
        m_data.mdoList.insert(m_data.mdoList.begin() + pos, cpMdo);
    }

    cpMdo->SetParent(this); // Adoptation

    // Sauvegarde du mdo
    cpMdo->SaveToCfg();

    // Refresh
    CreateMdoListStr();
    SaveToCfg();
}
//------------------------------------------------------------------------------
void Sizer::UnlinkMdo(const DrawableObject *mdo) // Remove Mdo from list but do not destroy it
{
    m_data.mdoList.erase(m_data.mdoList.begin() + GetMdoPos(mdo));

    CreateMdoListStr();
    SaveToCfg();
}
//------------------------------------------------------------------------------
int Sizer::GetMdoPos(const DrawableObject *mdo) const
{
    for(int i = 0; i < m_data.mdoList.size(); i++) // Pas de "break;" par plusieurs pointeur possible.
    {
        if(mdo == m_data.mdoList[i])
        {
            return i;
        }
    }

    throw ticpp::Exception("Sizer::GetMdoPos : DrawableObject not found !");
}
//------------------------------------------------------------------------------
DrawableObject* Sizer::AddMdo(const std::string &type, const std::string &name)
{
    if(type == MDO_STATIC_TEXT_TYPE_NAME)
    {
        m_data.mdoList.push_back(new StaticText(m_cfg, m_devisCfg, this, name));
    }
    else if(type == MDO_DEVIS_TEXT_TYPE_NAME)
    {
        m_data.mdoList.push_back(new DevisText(m_cfg, m_devisCfg, this, name));
    }
    else if(type == MDO_IMAGE_TYPE_NAME)
    {
        m_data.mdoList.push_back(new Image(m_cfg, m_devisCfg, this, name));
    }
    else if(type == MDO_SIZER_TYPE_NAME)
    {
        m_data.mdoList.push_back(new Sizer(m_cfg, m_devisCfg, this, name));
    }
    else if(type == MDO_SPACER_TYPE_NAME)
    {
        m_data.mdoList.push_back(new Spacer(m_cfg, m_devisCfg, this, name));
    }
    else
    {
        std::string msg = "mdo::Sizer AddMdo ==> Unknown mdo type \"" + type + "\".";
        throw ticpp::Exception(msg.c_str());
    }

    return m_data.mdoList[m_data.mdoList.size() - 1];
    SaveToCfg();
}
//------------------------------------------------------------------------------
DrawableObject* Sizer::CreateMdo(const std::string &type)
{
    DrawableObject *mdo = AddMdo(type, "name");
    mdo->AutoChangeConfigName();
    mdo->CreateToCfg();
    CreateMdoListStr();
    SaveToCfg();
    return mdo;
}
//------------------------------------------------------------------------------
DrawableObject* Sizer::LoadMdo(const std::string &type, const std::string &name)
{
    DrawableObject *mdo = AddMdo(type, name);
    mdo->LoadFromCfg();
    return mdo;
}
//------------------------------------------------------------------------------
int Sizer::GetMdoCount() const
{
    return m_data.mdoList.size();
}
//------------------------------------------------------------------------------
DrawableObject* Sizer::GetMdoPtr(const int &pos)
{
    return m_data.mdoList[pos];
}
//------------------------------------------------------------------------------
void Sizer::CreateMdoListStr()
{
    m_data.mdoStrList = "";
    for(int i = 0; i < m_data.mdoList.size(); i++)
    {
        m_data.mdoStrList += m_data.mdoList[i]->GetType() + std::string(":") + m_data.mdoList[i]->GetConfigName() + std::string(";");
    }
}
//------------------------------------------------------------------------------
bool Sizer::SearchTreeItemId(const wxTreeItemId &treeItemId, DrawableObject *&mdo)
{
    // Test sur l'objet actuel : ne peut être ok que si c'est le root sizer
    if(GetTreeItemId() == treeItemId)
    {
        mdo = this;
        return true;
    }

    // Recherche dans tous les enfants...
    for(int i = 0; i < m_data.mdoList.size(); i++)
    {
        // Est-ce cet enfant ?
        if(m_data.mdoList[i]->GetTreeItemId() == treeItemId)
        {
            mdo = m_data.mdoList[i];
            return true;
        }

        if(m_data.mdoList[i]->GetType() == MDO_SIZER_TYPE_NAME) // Un sizer ==> Recherche à l'interieur
        {
            Sizer *sizer = static_cast<Sizer*>(m_data.mdoList[i]);

            if(sizer->SearchTreeItemId(treeItemId, mdo))
            {
                // Si il fait parti du sizer (ou de ses sous sizer)
                return true; // ok
            }
        }
    }

    return false;
}
//------------------------------------------------------------------------------
bool Sizer::IsVertical() const
{
    return m_data.isVertical;
}
//------------------------------------------------------------------------------
bool Sizer::IsRoot() const
{
    return m_parentSizer == NULL;
}
//------------------------------------------------------------------------------

} // namespace mdo

